@@ -78,6 +78,7 @@ fn on_struct(inp: &mut syn::DeriveInput) -> syn::Result<proc_macro2::TokenStream
 
     let field_str  = fields.idents.iter().map(|n| format!("{}::{}", name, n)).collect::<Vec<_>>();
     let statements = gen_statements(&fields, &decode_fns, attrs.encoding().unwrap_or_default())?;
+    let nils       = nils(&fields.types, &decode_fns);
 
     let Fields { indices, idents, .. } = fields;
 
@@ -86,8 +87,10 @@ fn on_struct(inp: &mut syn::DeriveInput) -> syn::Result<proc_macro2::TokenStream
             Ok(#name {
                 #(#idents : if let Some(x) = #idents {
                     x
+                } else if let Some(z) = #nils {
+                    z
                 } else {
-                    return Err(minicbor::decode::Error::MissingValue(#indices, #field_str))
+                    return Err(minicbor::decode::Error::missing_value(#indices).with_message(#field_str).at(__p777))
                 }),*
             })
         }
@@ -97,8 +100,10 @@ fn on_struct(inp: &mut syn::DeriveInput) -> syn::Result<proc_macro2::TokenStream
         quote! {
             Ok(#name(#(if let Some(x) = #idents {
                 x
+            } else if let Some(z) = #nils {
+                z
             } else {
-                return Err(minicbor::decode::Error::MissingValue(#indices, #field_str))
+                return Err(minicbor::decode::Error::missing_value(#indices).with_message(#field_str).at(__p777))
             }),*))
         }
     };
@@ -106,6 +111,7 @@ fn on_struct(inp: &mut syn::DeriveInput) -> syn::Result<proc_macro2::TokenStream
     Ok(quote! {
         impl #impl_generics minicbor::Decode<'bytes> for #name #typ_generics #where_clause {
             fn decode(__d777: &mut minicbor::Decoder<'bytes>) -> core::result::Result<#name #typ_generics, minicbor::decode::Error> {
+                let __p777 = __d777.position();
                 #statements
                 #result
             }
@@ -166,6 +172,7 @@ fn on_enum(inp: &mut syn::DeriveInput) -> syn::Result<proc_macro2::TokenStream>
                 collect_type_params(&inp.generics, iter)
             });
             let statements = gen_statements(&fields, &decode_fns, encoding)?;
+            let nils       = nils(&fields.types, &decode_fns);
             let Fields { indices, idents, .. } = fields;
             if let syn::Fields::Named(_) = var.fields {
                 quote! {
@@ -174,8 +181,10 @@ fn on_enum(inp: &mut syn::DeriveInput) -> syn::Result<proc_macro2::TokenStream>
                         Ok(#name::#con {
                             #(#idents : if let Some(x) = #idents {
                                 x
+                            } else if let Some(z) = #nils {
+                                z
                             } else {
-                                return Err(minicbor::decode::Error::MissingValue(#indices, #field_str))
+                                return Err(minicbor::decode::Error::missing_value(#indices).with_message(#field_str).at(__p777))
                             }),*
                         })
                     }
@@ -186,8 +195,10 @@ fn on_enum(inp: &mut syn::DeriveInput) -> syn::Result<proc_macro2::TokenStream>
                         #statements
                         Ok(#name::#con(#(if let Some(x) = #idents {
                             x
+                        } else if let Some(z) = #nils {
+                            z
                         } else {
-                            return Err(minicbor::decode::Error::MissingValue(#indices, #field_str))
+                            return Err(minicbor::decode::Error::missing_value(#indices).with_message(#field_str).at(__p777))
                         }),*))
                     }
                 }
@@ -208,12 +219,16 @@ fn on_enum(inp: &mut syn::DeriveInput) -> syn::Result<proc_macro2::TokenStream>
     let (_, typ_generics, where_clause) = inp.generics.split_for_impl();
 
     let check = if index_only {
-        quote!()
+        quote! {
+            let __p778 = __d777.position();
+        }
     } else {
         quote! {
+            let __p777 = __d777.position();
             if Some(2) != __d777.array()? {
-                return Err(minicbor::decode::Error::Message("expected enum (2-element array)"))
+                return Err(minicbor::decode::Error::message("expected enum (2-element array)").at(__p777))
             }
+            let __p778 = __d777.position();
         }
     };
 
@@ -223,7 +238,7 @@ fn on_enum(inp: &mut syn::DeriveInput) -> syn::Result<proc_macro2::TokenStream>
                 #check
                 match __d777.u32()? {
                     #(#rows)*
-                    n => Err(minicbor::decode::Error::UnknownVariant(n))
+                    n => Err(minicbor::decode::Error::unknown_variant(n).at(__p778))
                 }
             }
         }
@@ -264,27 +279,50 @@ fn gen_statements(fields: &Fields, decode_fns: &[Option<CustomCodec>], encoding:
     let actions = fields.indices.iter().zip(fields.idents.iter().zip(fields.types.iter().zip(decode_fns)))
         .map(|(ix, (name, (ty, ff)))| {
             let decode_fn = ff.as_ref()
-                .and_then(|ff| ff.to_decode_path())
+                .and_then(CustomCodec::to_decode_path)
                 .unwrap_or_else(|| default_decode_fn.clone());
-            if is_option(ty, |_| true) {
-                return quote! {
-                    match #decode_fn(__d777) {
-                        Ok(__v777) => #name = Some(__v777),
-                        Err(minicbor::decode::Error::UnknownVariant(_)) => { __d777.skip()? }
-                        Err(e) => return Err(e)
+
+            let unknown_var_err =
+                if let Some(cd) = ff {
+                    if let Some(p) = cd.to_nil_path() {
+                        quote! {
+                            Err(e) if e.is_unknown_variant() && #p().is_some() => {
+                                __d777.skip()?
+                            }
+                        }
+                    } else if is_option(ty, |_| true) {
+                        quote! {
+                            Err(e) if e.is_unknown_variant() => __d777.skip()?,
+                        }
+                    } else {
+                        quote!()
                     }
+                } else if is_option(ty, |_| true) {
+                    quote! {
+                        Err(e) if e.is_unknown_variant() => __d777.skip()?,
+                    }
+                } else {
+                    quote! {
+                        Err(e) if e.is_unknown_variant() && <#ty>::nil().is_some() => {
+                            __d777.skip()?
+                        }
+                    }
+                };
+
+            let value =
+                if ix.is_b() && is_cow(ty, |t| is_str(t) || is_byte_slice(t)) {
+                    quote!(Some(minicbor::bytes::Cow::Borrowed(__v777)))
+                } else {
+                    quote!(Some(__v777))
+                };
+
+            quote! {
+                match #decode_fn(__d777) {
+                    Ok(__v777) => #name = #value,
+                    #unknown_var_err
+                    Err(e) => return Err(e)
                 }
             }
-            if ix.is_b() && is_cow(ty, |t| is_str(t) || is_byte_slice(t)) {
-                return quote! {
-                    match #decode_fn(__d777) {
-                        Ok(__v777) => #name = Some(minicbor::bytes::Cow::Borrowed(__v777)),
-                        Err(minicbor::decode::Error::UnknownVariant(_)) => { __d777.skip()? }
-                        Err(e) => return Err(e)
-                    }
-                }
-            }
-            quote!({ #name = Some(#decode_fn(__d777)?) })
     })
     .collect::<Vec<_>>();
 
@@ -375,3 +413,21 @@ fn make_transparent_impl
 fn gen_decode_bound() -> syn::Result<syn::TypeParamBound> {
     syn::parse_str("minicbor::Decode<'bytes>")
 }
+
+fn nils(types: &[syn::Type], decode_fns: &[Option<CustomCodec>]) -> Vec<proc_macro2::TokenStream> {
+    types.iter().zip(decode_fns)
+        .map(|(ty, dec)| {
+            if let Some(d) = dec {
+                if let Some(p) = d.to_nil_path() {
+                    quote!(#p())
+                } else if is_option(ty, |_| true) {
+                    quote!(Some(None))
+                } else {
+                    quote!(None)
+                }
+            } else {
+                quote!(<#ty>::nil())
+            }
+        })
+        .collect()
+}
